import sys
import json
from datetime import datetime

import requests


class LogEntry:
    """Class describing a connection attempt"""

    def __init__(self: 'LogEntry', log_entry: str) -> None:
        self.date = datetime.strptime(log_entry[1:20], '%Y-%m-%d %H:%M:%S')
        self.host = log_entry.split('[')[2].split(']')[0]
        self.failed = 'FAIL' in log_entry


def printOutages(config: dict) -> None:
    """Print outages from a log file"""

    with open(config['log'], 'r') as file:
        log = list(map(lambda x: LogEntry(x), file.readlines()))

    # cluster log entries by host
    log_by_host = {}
    for entry in log:
        if entry.host not in log_by_host:
            log_by_host[entry.host] = [entry]
        else:
            log_by_host[entry.host].append(entry)

    # cluster fails to outages
    # we consider the time of subsequent failed connection attemts to the same host an outage
    outages: list = []
    last_failed = False
    for entries in log_by_host.values():
        for entry in entries:
            if entry.failed and last_failed:
                outages[-1].append(entry)
            if entry.failed and not last_failed:
                outages.append([entry])
                last_failed = True
            if not entry.failed and last_failed:
                last_failed = False

    # print the outages by date of first fail
    for outage in sorted(outages, key=lambda x: x[0].date):

        # get duration of outage in hours and minutes
        outage_duration = round(
            (outage[-1].date - outage[0].date).seconds / 60)
        # continue if outage is shorter than the minimum duration
        if outage_duration < config['outages']:
            continue
        hours = outage_duration // 60
        minutes = outage_duration - (hours * 60)
        host = outage[0].host

        # Outputs outages in the form of "Outage at: 2023-19-01 06:29:01 lasting for 2 Hours and 39 Minutes"
        print(f"[{outage[0].date}][{host}] lasting for {'{} Hours and '.format(hours) if hours >= 1 else ''}{minutes} Minutes")


def isOnline(host: str, timeout: int) -> bool:
    """Check if connection to a host can be established"""

    try:
        requests.head(f"https://{host}", timeout=timeout)
        return True
    except requests.ConnectionError:
        return False


def log(config: dict) -> None:
    """Log the connection status of a host"""

    logline = f"[{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}][{config['host']}][{'OK' if isOnline(config['host'], config['timeout']) else 'FAIL'}]\n"

    if config['log'] != None:
        with open(config['log'], 'a') as file:
            file.write(logline)

    if config['stdout']:
        print(logline)


def printHelp() -> None:
    """Prints the help menu"""

    print(
        """
        If no argument is passed then de defaults are assumed.

        --config    Path to a config file in JSON format.
                    To set a command line argument, use it as a key and set its value.
                    Config is nor used unless explicitly set.

        --log       Path to a the file where results should be appended to.
                    Created the file if it does not already exist.
                    Default: ./connection.log

        --host      IP or hostname of a server to query.
                    Default: 1.1.1.1

        --timeout   Set the timeout in seconds to use for the connection test
                    Default: 2

        --outages   Print the outages to stdout.
                    Set the lenght in minutes where a connection loss is condidered an outage
                    This option can only be used with --log.
                    Default: 3

        --stdout    Return the resulting logline in the terminal.
                    Default Behaviour: Do not print to stdout.

        --help      Print this menu
        """
    )


def loadConfig() -> tuple[dict, str]:
    """Load and validate the config and return it as a dict together with the execution mode"""

    argv = sys.argv

    config: dict = {}
    buffer = ''
    mode = 'log'

    # parse arguments
    for arg in argv[1:]:
        if arg.startswith('--'):
            buffer = arg.removeprefix('--')
            config[buffer] = []
        else:
            config[buffer].append(arg)

    # print help if requested
    if 'help' in config:
        printHelp()
        quit()

    ## Validate Arguments ##

    # right arguments
    valid_parameters = {'config', 'log', 'host',
                        'timeout', 'outages', 'stdout', 'help'}
    if (unknown_parameters := config.keys() - valid_parameters) != set():
        print(f"[ERROR] Unknown parameters: {unknown_parameters}")
        printHelp()
        quit()

    # number of arguments
    maximum_arguments = {
        'config': 1,
        'log': 1,
        'host': 1,
        'timeout': 1,
        'outages': 1,
        'stdout': 0,
        'help': 0
    }

    for key in config:
        if len(config[key]) > maximum_arguments[key]:
            print(
                f"[ERROR] Too many arguments for '{key}'. Expected 0{' or 1' if maximum_arguments[key] == 1 else ''}. Got: {len(config[key])}")
            printHelp()
            quit()

    # check that outages is only used with --log
    if 'outages' in config:
        mode = 'outages'
        if config.keys() - {'outages', 'log'} != set():
            print(
                f"[ERROR] --outages can only be used alone or with --log")
            printHelp()
            quit()

    # After that check has passed remove the list
    for key in config:
        config[key] = config[key][0] if len(config[key]) == 1 else None

    # set stdout
    config['stdout'] = 'stdout' in config

    # if config is passed then load it
    if 'config' in config:
        try:
            config_file = './config.json' if config['config'] == None else config['config']
            with open(config_file, 'r') as file:
                # this gives the stdout toggle behaviour
                config_file = json.load(file)

                # merge config_file and config
                # this also subtly overrites no-stdout to whatever the config file says
                for key in config_file.keys() & config.keys():
                    if key == 'stdout':
                        continue
                    if config[key] == None:
                        config[key] = config_file[key]

                config_file.update(config)
                config = config_file

        except FileNotFoundError:
            print(f"[ERROR] Config file not found: {config['config']}")
            printHelp()
            quit()

    # load defaults, if not set
    defaults = {
        'log': './connection.log',
        'host': '1.1.1.1',
        'timeout': 2,
        'outages': 3,
        'stdout': False,
        'help': None
    }
    for key in defaults:
        if key == 'log':
            if config['stdout'] and 'log' not in config:
                config['log'] = None
                continue

        if (key not in config) or (key in config and config[key] == None):
            config[key] = defaults[key]

    # timeout
    try:
        config['timeout'] = int(config['timeout'])
    except ValueError:
        print(
            f"[ERROR] Invalid type for 'timeout'. Expected int. Got: {type(config['timeout'])}")
        printHelp()
        quit()

    if config['timeout'] <= 0:
        print(f"[ERROR] 'timeout' needs to be larger than 0")
        printHelp()
        quit()

    # outages
    try:
        config['outages'] = int(config['outages'])
    except ValueError:
        print(
            f"[ERROR] Invalid type for 'outages'. Expected int. Got: {type(config['outages'])}")
        printHelp()
        quit()

    if config['outages'] < 0:
        print(f"[ERROR] 'outages' needs to be larger than 0")
        printHelp()
        quit()

    return config, mode


def main():
    config, mode = loadConfig()
    if mode == 'log':
        log(config)
    elif mode == 'outages':
        printOutages(config)


if __name__ == '__main__':
    main()
